public class NTPMessage {
    private byte leap, version, mode;
    private byte stratum, poll;
    private byte precision;
    private int delay;
    private int dispersion;
    private byte[] reference;
    private Timestamp referenceT, originateT, receiveT, transmitT;

    public NTPMessage(byte l, byte v, byte m, byte s, byte pl, byte pr,
                      int del, int disp, byte[] r,
                      Timestamp rft, Timestamp ot,
                      Timestamp rct, Timestamp tt) {
        if(r != null && r.length != 4)
            throw new Error("Reference doit avoir quatre caractres!");
        if(del > 0x100000000L)
            throw new Error("Delay doit tre sur 32 bits");
        if(disp > 0x100000000L)
            throw new Error("Dispersion doit tre sur 32 bits");
        if(pr >= 128 || pr < -128)
            throw new Error("Precision doit tre sur 8 bits");
        leap = l;
        version = v;
        mode = m;
        stratum = s;
        poll = pl;
        precision = pr;
        delay = del;
        dispersion = disp;
        reference = r;
        referenceT = (Timestamp)rft.clone();
        originateT = (Timestamp)ot.clone();
        receiveT = (Timestamp)rct.clone();
        transmitT = (Timestamp)tt.clone();
    }

    public NTPMessage(Timestamp t) {
        this((byte)0,           // pas de seconde ajoute
             (byte)3,           // version 3
             (byte)3,           // mode client
             (byte)0,
             (byte)0,
             (byte)0,
             0,                 // 1 s environ
             0,                 // 50 ms environ
             null,              // inconnu
             new Timestamp(0L), // indfini
             new Timestamp(0L),
             new Timestamp(0L),
             t);
    }
            

    public NTPMessage(byte[] buf, int offset) {
        leap = (byte)(buf[offset] >> 6);
        version = (byte)((buf [offset] >> 3) & 0x5);
        mode = (byte)(buf[offset] & 0x5);
        stratum = buf[offset + 1];
        poll = buf[offset + 2];
        precision = buf[offset + 3];
        delay = getS32(buf, 4);
        dispersion = getS32(buf, 8);
        reference = new byte[4];
        reference[0] = buf[12];
        reference[1] = buf[13];
        reference[2] = buf[14];
        reference[3] = buf[15];
        referenceT = new Timestamp(buf, 16);
        originateT = new Timestamp(buf, 24);
        receiveT = new Timestamp(buf, 32);
        transmitT = new Timestamp(buf, 40);
    }
        
    public NTPMessage(byte[] buf) {
        this(buf, 0);
    }

    public void format(byte[] buf, int offset) {
        buf[0] = (byte)(leap << 6 | version << 3 | mode);
        buf[1] = stratum;
        buf[2] = poll;
        buf[3] = (byte)(precision & 0xFF);
        format32(delay, buf, 4);
        format32(dispersion, buf, 8);
        if(reference != null) {
            buf[12] = reference[0];
            buf[13] = reference[1];
            buf[14] = reference[2];
            buf[15] = reference[3];
        } else {
            buf[12] = buf[13] = buf[14] = buf[15] = 0;
        }
        referenceT.format(buf, 16);
        originateT.format(buf, 24);
        receiveT.format(buf, 32);
        transmitT.format(buf, 40);
    }

    public void format(byte[] buf) {
        format(buf, 0);
    }

    public byte[] format() {
        byte[] buf;
        buf = new byte[48];
        format(buf);
        return buf;
    }

    public String toString() {
        return "NTPMessage " +
            originateT + " -> " + receiveT + ", " + transmitT + 
            " (" + referenceT + ")";
    }

    private static long get32(byte[] buf, int offset) {
        return 
            ((((buf[offset] & 0xFF) * 0x100L) +
              (buf[offset + 1] & 0xFF)) * 0x100L +
             (buf[offset + 2] & 0xFF)) * 0x100L +
            (buf[offset + 3] & 0xFF);
    }

    private static int getS32(byte[] buf, int offset) {
        return (int)(get32(buf, offset) & 0xFFFFFFFFL);
    }

    private static void format32(long v, byte[] buf, int offset) {
        if(v < 0 || v >= 0x100000000L)
            throw new Error("Format 32 fonctionne sur 32 bits");
        buf[offset] = (byte)(v >> 24);
        buf[offset + 1] = (byte)((v >> 16) & 0xFF);
        buf[offset + 2] = (byte)((v >> 8) & 0xFF);
        buf[offset + 3] = (byte)(v & 0xFF);
    }

    private static void formatS32(int v, byte[] buf, int offset) {
        long vv = (long)v & 0xFFFFFFFFL;
        format32(vv, buf, offset);
    }
}
