import java.util.Date;

public class Timestamp implements Cloneable {
    private long seconds, fraction;

    public Timestamp(long s, long f) {
        seconds = s;
        fraction = f;
    }
    
    public Timestamp(byte[] buf, int offset) {
        seconds = get32(buf, offset);
        fraction = get32(buf, offset + 4);
    }

    public Timestamp(byte[] buf) {
        this(buf, 0);
    }

    public void format(byte[] buf, int offset) {
        format32(seconds, buf, offset);
        format32(fraction, buf, offset + 4);
    }

    public Timestamp(double d) {
        seconds = (long)d;
        fraction = (long)(d * (double)0x100000000L);
    }

    public Timestamp(float d) {
        seconds = (long)d;
        fraction = (long)((double)d * (double)0x100000000L);
    }

    public static Timestamp fromDate(long t) {
        long s, f;
        s = (t / 1000L) + 0x83aa7e80L;
        f = (t % 1000L) * 1000L * 4295;
        return new Timestamp(s, f);
    }

    public static Timestamp fromDate(Date d) {
        return fromDate(d.getTime());
    }

    public static Timestamp currentTime() {
        return fromDate(System.currentTimeMillis());
    }

    public long toMillis() {
        long s, ms;
        s = seconds - 0x83aa7e80L;
        ms = (long)((double)fraction / (double)1000L / (double)4295 + 0.5);
        return s * 1000L + ms;
    }

    public Date toDate() {
        return new Date(toMillis());
    }

    public long diffMillis(Timestamp since) {
        return 
            (seconds - since.seconds) * 1000L +
            (fraction - since.fraction) / 1000L / 4295L;
    }

    public String toString() {
        if(seconds == 0)
            return "[??]";
        else
            return "[" + toDate() + " " + fraction / 1000L / 4295L + "]";
    }

    public float toFloat() {
        return (float)seconds + 
            (float)fraction / (float)0x100000000L;
    }

    public double toDouble() {
        return (double)seconds +
            (double)fraction / (double)0x100000000L;
    }

    public Object clone() {
        try {
            return super.clone();
        } catch(CloneNotSupportedException e) {
            throw new Error("Ceci ne peut pas arriver " + e);
        }
    }

    private static long get32(byte[] buf, int offset) {
        return 
            ((((buf[offset] & 0xFF) * 0x100L) +
              (buf[offset + 1] & 0xFF)) * 0x100L +
             (buf[offset + 2] & 0xFF)) * 0x100L +
            (buf[offset + 3] & 0xFF);
    }

    private static void format32(long v, byte[] buf, int offset) {
        if(v < 0 || v >= 0x100000000L)
            throw new Error("Format 32 fonctionne sur 32 bits");
        buf[offset] = (byte)(v >> 24);
        buf[offset + 1] = (byte)((v >> 16) & 0xFF);
        buf[offset + 2] = (byte)((v >> 8) & 0xFF);
        buf[offset + 3] = (byte)(v & 0xFF);
    }

}
